"""
pyEQL osmotic coefficient test suite
============================================

This file contains tests for the osmotic coefficient method
employed by pyEQL.

NOTE: generally, these tests check the module output against experimental
data rather than the theoretical result of the respective functions.
"""

import numpy as np

from pyEQL import Solution


def test_osmotic_pressure():
    """
    The osmotic pressure of seawater is approximately 27 atm
    """
    # TODO - at present this test is inaccurate because in the complex matrix
    # of seawater, pyEQL falls back to using an ideal solution model with
    # unit osmotic coefficient.
    empty = Solution()
    assert np.isclose(empty.osmotic_pressure.to("atm").magnitude, 0, atol=1e-5)
    sea = Solution.from_preset("seawater")
    assert np.isclose(sea.osmotic_pressure.to("atm").magnitude, 27, rtol=0.15)


class Test_osmotic_pitzer:
    """
    test osmotic coefficient based on the Pitzer model
    ------------------------------------------------

    """

    def test_dimensionality(self):
        s1 = Solution([["Na+", "0.1 mol/L"], ["Cl-", "0.1 mol/L"]])
        assert s1.get_osmotic_coefficient().dimensionality == ""
        assert s1.get_osmotic_coefficient() >= 0

    def test_osmotic_pitzer_ammoniumnitrate(self):
        """
        calculate the osmotic coefficient at each concentration and compare
        to experimental data for ammonium nitrate

        References
        ----------
        May, P. M., Rowland, D., Hefter, G., & Königsberger, E. (2011).
        A Generic and Updatable Pitzer Characterization of Aqueous Binary Electrolyte Solutions at 1 bar and 25 °C.
        Journal of Chemical & Engineering Data, 56(12), 5066-5077. doi:10.1021/je2009329

        """
        # list of concentrations to test, mol/kg
        conc_list = [0.25, 0.5, 0.75, 1, 1.5, 2]

        # list of published experimental osmotic coefficients
        pub_osmotic_coeff = [0.86, 0.855, 0.83, 0.825, 0.80, 0.78]

        for i, conc in enumerate(conc_list):
            conc = str(conc) + "mol/kg"
            sol = Solution()
            sol.add_solute("NH4+", conc)
            sol.add_solute("NO3-", conc)
            result = sol.get_osmotic_coefficient()
            expected = pub_osmotic_coeff[i]

            assert np.isclose(result, expected, rtol=0.05)

    def test_osmotic_pitzer_coppersulfate(self):
        """
        calculate the osmotic coefficient at each concentration and compare
        to experimental data for copper sulate

        References
        ----------
        May, P. M., Rowland, D., Hefter, G., & Königsberger, E. (2011).
        A Generic and Updatable Pitzer Characterization of Aqueous Binary Electrolyte Solutions at 1 bar and 25 °C.
        Journal of Chemical & Engineering Data, 56(12), 5066-5077. doi:10.1021/je2009329

        """
        # list of concentrations to test, mol/kg
        conc_list = [0.25, 0.5, 0.75, 1]

        # list of published experimental osmotic coefficients
        pub_osmotic_coeff = [0.5, 0.485, 0.48, 0.485, 0.5]

        for i, conc in enumerate(conc_list):
            conc = str(conc) + "mol/kg"
            sol = Solution()
            sol.add_solute("Cu+2", conc)
            sol.add_solute("SO4-2", conc)
            result = sol.get_osmotic_coefficient()
            expected = pub_osmotic_coeff[i]

            assert np.isclose(result, expected, rtol=0.05)



"""
Tests for the solute.py module
"""
from pyEQL.solute import Solute


def test_from_formula():
    s = Solute.from_formula("Mg+2")
    assert s.formula == "Mg[+2]"
    assert s.formula_pretty == "Mg^+2"
    assert s.formula_html == "Mg<sup>+2</sup>"
    assert s.formula_hill == "Mg"
    assert s.formula_latex == "Mg$^{+2}$"
    assert s.chemsys == "Mg"
    assert s.charge == 2
    assert s.n_atoms == 1
    assert s.n_elements == 1
    assert s.oxi_state_guesses == {"Mg": 2.0}
    # test behavior when oxi_state_guesses fails
    assert Solute.from_formula("Br[-0.33333333]").oxi_state_guesses == {}
    assert s.molecular_weight == "24.305 g/mol"
    s2 = Solute.from_formula("O6")
    assert s2.formula == "O3(aq)"
    assert s2.molecular_weight == "47.9982 g/mol"
    assert s2.oxi_state_guesses == {"O": 0.0}



"""
pyEQL functions that take Solution objects as inputs or return Solution objects.

:copyright: 2013-2024 by Ryan S. Kingsbury
:license: LGPL, see LICENSE for more details.

"""

import logging

import numpy as np

from pyEQL import Solution, ureg

logger = logging.getLogger(__name__)


def gibbs_mix(solution1: Solution, solution2: Solution):
    r"""
    Return the Gibbs energy change associated with mixing two solutions.

    Args:
        solution1: a solution to be mixed.
        solution2: a solution to be mixed.

    Returns:
        The change in Gibbs energy associated with complete mixing of the
        Solutions, in Joules.

    Notes:
        The Gibbs energy of mixing is calculated as follows

        .. math::

            \Delta_{mix} G = \sum_i {(n_c + n_d) R T \ln a_b} - \sum_i {n_c R T \ln a_c} - \sum_i {n_d R T \ln a_d}

        Where :math:`n` is the number of moles of substance, :math:`T` is the temperature in kelvin,
        and  subscripts :math:`b`, :math:`c`, and :math:`d` refer to the concentrated, dilute, and blended
        Solutions, respectively.

        Note that dissociated ions must be counted as separate components,
        so a simple salt dissolved in water is a three component solution (cation,
        anion, and water).

    References:
        Koga, Yoshikata, 2007. Solution Thermodynamics and its Application to Aqueous Solutions:
            A differential approach. Elsevier, 2007, pp. 23-37.

    """
    concentrate = solution1
    dilute = solution2
    blend = solution1 + solution2
    term_list = {concentrate: 0, dilute: 0, blend: 0}

    # calculate the entropy change and number of moles solute for each solution
    for solution in term_list:
        for solute in solution.components:
            if solution.get_amount(solute, "fraction") != 0:
                term_list[solution] += solution.get_amount(solute, "mol") * np.log(solution.get_activity(solute))

    return (ureg.R * blend.temperature.to("K") * (term_list[blend] - term_list[concentrate] - term_list[dilute])).to(
        "J"
    )


def entropy_mix(solution1: Solution, solution2: Solution):
    r"""
    Return the ideal mixing entropy associated with mixing two solutions.

    Parameters:
        solution1, solution2: The two solutions to be mixed.

    Returns:
        The ideal mixing entropy associated with complete mixing of the
        Solutions, in Joules.

    Notes:
        The ideal entropy of mixing is calculated as follows

        .. math::

            \Delta_{mix} S = \sum_i {(n_c + n_d) R T \ln x_b} - \sum_i {n_c R T \ln x_c} - \sum_i {n_d R T \ln x_d}

        Where :math:`n` is the number of moles of substance, :math:`T` is the temperature in kelvin,
        and  subscripts :math:`b`, :math:`c`, and :math:`d` refer to the concentrated, dilute, and blended
        Solutions, respectively.

        Note that dissociated ions must be counted as separate components,
        so a simple salt dissolved in water is a three component solution (cation,
        anion, and water).

    References:
        Koga, Yoshikata, 2007. *Solution Thermodynamics and its Application to Aqueous Solutions:
            A differential approach.* Elsevier, 2007, pp. 23-37.

    """
    concentrate = solution1
    dilute = solution2
    blend = solution1 + solution2
    term_list = {concentrate: 0, dilute: 0, blend: 0}

    # calculate the entropy change and number of moles solute for each solution
    for solution in term_list:
        for solute in solution.components:
            if solution.get_amount(solute, "fraction") != 0:
                term_list[solution] += solution.get_amount(solute, "mol") * np.log(
                    solution.get_amount(solute, "fraction")
                )

    return (ureg.R * blend.temperature.to("K") * (term_list[blend] - term_list[concentrate] - term_list[dilute])).to(
        "J"
    )


def donnan_eql(solution: Solution, fixed_charge: str):
    r"""
    Return a solution object in equilibrium with fixed_charge.

    Args:
        solution : Solution object
            The external solution to be brought into equilibrium with the fixed
            charges
        fixed_charge : str quantity
            String representing the concentration of fixed charges, including sign.
            May be specified in mol/L or mol/kg units. e.g. '1 mol/kg'

    Returns:
        A Solution that has established Donnan equilibrium with the external
        (input) Solution

    Notes:
        The general equation representing the equilibrium between an external
        electrolyte solution and an ion-exchange medium containing fixed charges
        is

        .. math::

            \big(\frac{a_{-}}{\bar a_{-}} \big)^(\frac{1}{z_{-})
            \big(\frac{\bar a_{+}}{a_{+}}\big)^(\frac{1}{z_{+})
            \exp(\frac{\Delta \pi \bar V}{RT z_{+} \nu_{+}})

        Where subscripts :math:`+` and :math:`-` indicate the cation and anion, respectively,
        the overbar indicates the membrane phase,
        :math:`a` represents activity, :math:`z` represents charge, :math:`\nu` represents the stoichiometric
        coefficient, :math:`V` represents the partial molar volume of the salt, and
        :math:`\Delta \pi` is the difference in osmotic pressure between the membrane and the
        solution phase.

        In addition, electroneutrality must prevail within the membrane phase:

        .. math:: \bar C_{+} z_{+} + \bar X + \bar C_{-} z_{-} = 0

        Where :math:`C` represents concentration and :math:`X` is the fixed charge concentration
        in the membrane or ion exchange phase.

        This function solves these two equations simultaneously to arrive at the
        concentrations of the cation and anion in the membrane phase. It returns
        a solution equal to the input solution except that the concentrations of
        the predominant cation and anion have been adjusted according to this
        equilibrium.

        NOTE that this treatment is only capable of equilibrating a single salt.
        This salt is identified by the get_salt() method.

    References:
        Strathmann, Heiner, ed. *Membrane Science and Technology* vol. 9, 2004. Chapter 2, p. 51.
           http://dx.doi.org/10.1016/S0927-5193(04)80033-0

    See Also:
        get_salt()

    """
    # identify the salt
    salt = solution.get_salt()

    # convert fixed_charge in to a quantity
    fixed_charge = ureg.Quantity(fixed_charge)

    # identify variables from the external solution
    conc_cation_soln = solution.get_amount(salt.cation, str(fixed_charge.units))
    conc_anion_soln = solution.get_amount(salt.anion, str(fixed_charge.units))
    act_cation_soln = solution.get_activity(salt.cation)
    act_anion_soln = solution.get_activity(salt.anion)
    z_cation = salt.z_cation
    z_anion = salt.z_anion
    nu_cation = salt.nu_cation

    # get the partial molar volume for the salt, or calculate it from the ions
    # TODO - consider how to incorporate pitzer parameters
    molar_volume = solution.get_property(salt.formula, "size.molar_volume")
    if molar_volume is None:
        cation_vol = solution.get_property(salt.cation, "size.molar_volume")
        anion_vol = solution.get_property(salt.anion, "size.molar_volume")
        if cation_vol is not None and anion_vol is not None:
            molar_volume = cation_vol + anion_vol
        else:
            logger.error("Required partial molar volume information not available. Aborting.")
            return None

    # initialize the equilibrated solution - start with a direct copy of the
    # input / external solution
    donnan_soln = solution.copy()

    # do nothing if either of the ion concentrations is zero
    if conc_cation_soln.magnitude == 0 or conc_anion_soln.magnitude == 0:
        return donnan_soln

    # define a function representing the donnan equilibrium as a function
    # of the two unknown activities to feed to the nonlinear solver

    # the stuff in the term below doesn't change on iteration, so calculate it up-front
    # assign it the correct units and extract the magnitude for a performance gain
    exp_term = (molar_volume / (ureg.R * solution.temperature * z_cation * nu_cation)).to("1/Pa").magnitude

    def donnan_solve(x):
        """Where x is the magnitude of co-ion concentration."""
        # solve for the counter-ion concentration by enforcing electroneutrality
        # using only floats / ints here instead of quantities helps performance
        if fixed_charge.magnitude >= 0:
            # counter-ion is the anion
            conc_cation_mem = x / abs(z_cation)
            conc_anion_mem = -(conc_cation_mem * z_cation + fixed_charge.magnitude) / z_anion
        elif fixed_charge.magnitude < 0:
            # counter-ion is the cation
            conc_anion_mem = x / abs(z_anion)
            conc_cation_mem = -(conc_anion_mem * z_anion + fixed_charge.magnitude) / z_cation

        # match the units given for fixed_charge
        units = str(fixed_charge.units)

        # set the cation and anion concentrations in the membrane phase equal
        # to the current guess
        donnan_soln.set_amount(salt.cation, str(conc_cation_mem) + units)
        donnan_soln.set_amount(salt.anion, str(conc_anion_mem) + units)

        # get the new concentrations and activities
        act_cation_mem = donnan_soln.get_activity(salt.cation)
        act_anion_mem = donnan_soln.get_activity(salt.anion)

        # compute the difference in osmotic pressure
        # using the magnitudes here helps performance
        delta_pi = donnan_soln.osmotic_pressure.magnitude - solution.osmotic_pressure.magnitude

        return (act_cation_mem / act_cation_soln) ** (1 / z_cation) * (act_anion_soln / act_anion_mem) ** (
            1 / z_anion
        ) - np.exp(delta_pi * exp_term)

    # solve the function above using one of scipy's nonlinear solvers

    from scipy.optimize import brentq

    # determine which ion concentration represents the co-ion
    # call a nonlinear solver to adjust the concentrations per the donnan
    # equilibrium, unless the membrane is uncharged
    # the initial guess is to set the co-ion concentration in the membrane
    # equal to that in the solution
    if fixed_charge.magnitude > 0:
        x = conc_cation_soln.magnitude
        brentq(donnan_solve, 1e-10, x, xtol=0.001)
    elif fixed_charge.magnitude < 0:
        x = conc_anion_soln.magnitude
        brentq(donnan_solve, 1e-10, x, xtol=0.001)
    else:
        pass

    # return the equilibrated solution
    return donnan_soln



"""
Tests of pyEQL.equilibrium module

"""
import numpy as np

from pyEQL.equilibrium import alpha


def test_alpha():
    # for each case, test pH << pKa, pH=pKa, pH >> pKa

    # monoprotic acid
    assert np.isclose(alpha(0, 3, [7]), 1, atol=1e-3)
    assert np.isclose(alpha(1, 3, [7]), 0, atol=1e-3)

    assert np.isclose(alpha(0, 7, [7]), 0.5, atol=1e-3)
    assert np.isclose(alpha(1, 7, [7]), 0.5, atol=1e-3)

    assert np.isclose(alpha(0, 10, [7]), 0, atol=1e-3)
    assert np.isclose(alpha(1, 10, [7]), 1, atol=1e-3)

    # carbonic acid, diprotic
    assert np.isclose(alpha(0, 3, [6.35, 10.33]), 1, atol=1e-3)
    assert np.isclose(alpha(1, 3, [6.35, 10.33]), 0, atol=1e-3)
    assert np.isclose(alpha(2, 3, [6.35, 10.33]), 0, atol=1e-3)

    assert np.isclose(alpha(0, 6.35, [6.35, 10.33]), 0.5, atol=1e-3)
    assert np.isclose(alpha(1, 6.35, [6.35, 10.33]), 0.5, atol=1e-3)

    assert np.isclose(alpha(0, 7.35, [6.35, 10.33]), 0.1, atol=1e-2)
    # test sorting
    assert np.isclose(alpha(0, 7.35, [10.33, 6.35]), 0.1, atol=1e-2)
    assert np.isclose(alpha(1, 7.35, [6.35, 10.33]), 0.9, atol=1e-2)
    assert np.isclose(alpha(2, 7.35, [6.35, 10.33]), 0, atol=1e-3)

    assert np.isclose(alpha(1, 10.33, [6.35, 10.33]), 0.5, atol=1e-3)
    assert np.isclose(alpha(2, 10.33, [6.35, 10.33]), 0.5, atol=1e-3)

    assert np.isclose(alpha(0, 12.33, [6.35, 10.33]), 0, atol=1e-3)
    assert np.isclose(alpha(1, 12.33, [6.35, 10.33]), 0.01, atol=1e-2)
    assert np.isclose(alpha(2, 12.33, [6.35, 10.33]), 0.99, atol=1e-2)



"""
pyEQL dielectric constant test suite
============================================

This file contains tests that check the dielectric constant
computations of pyEQL

"""

import numpy as np
import pyEQL
import pytest

# relative tolerance between experimental and computed properties for this test file
RTOL = 0.04


class Test_dielectric:
    """
    test the Dielectric Constant calculations of various solutions
    ------------------------------------------------

    Reference: A. Zuber, L. Cardozo-Filho, V.F. Cabral, R.F. Checoni, M. Castier, An empirical equation for the dielectric constant in aqueous and
    nonaqueous electrolyte mixtures, Fluid Phase Equilib. 376 (2014) 116-123. doi:10.1016/j.fluid.2014.05.037.

    """

    def test_dielectric_constant(self):
        """
        4.4 mol/kg NaCl = 46
        """
        s1 = pyEQL.Solution([["Na+", "4.4 mol/kg"], ["Cl-", "4.4 mol/kg"]])

        result = s1.dielectric_constant.magnitude
        expected = 46

        assert np.isclose(result, expected, rtol=RTOL)

    def test_dielectric_constant2(self):
        """
        2 mol/kg NaCl = 58
        """
        s1 = pyEQL.Solution([["Na+", "2 mol/kg"], ["Cl-", "2 mol/kg"]])

        result = s1.dielectric_constant.magnitude
        expected = 58

        assert np.isclose(result, expected, rtol=RTOL)

    def test_dielectric_constant3(self):
        """
        1 mol/kg NaCl = 66
        """
        s1 = pyEQL.Solution([["Na+", "1 mol/kg"], ["Cl-", "1 mol/kg"]])

        result = s1.dielectric_constant.magnitude
        expected = 66

        assert np.isclose(result, expected, rtol=RTOL)

    def test_dielectric_constant4(self):
        """
        1 mol/kg KBr = 67
        """
        s1 = pyEQL.Solution([["K+", "1 mol/kg"], ["Br-", "1 mol/kg"]])

        result = s1.dielectric_constant.magnitude
        expected = 67

        assert np.isclose(result, expected, rtol=RTOL)

    def test_dielectric_constant5(self):
        """
        3.4 mol/kg KBr = 51
        """
        s1 = pyEQL.Solution([["K+", "3.4 mol/kg"], ["Br-", "3.4 mol/kg"]])

        result = s1.dielectric_constant.magnitude
        expected = 51

        assert np.isclose(result, expected, rtol=RTOL)

    def test_dielectric_constant6(self):
        """
        5 mol/kg LiCl = 39
        """
        s1 = pyEQL.Solution([["Li+", "5 mol/kg"], ["Cl-", "5 mol/kg"]])

        result = s1.dielectric_constant.magnitude
        expected = 39

        assert np.isclose(result, expected, rtol=RTOL)

    def test_dielectric_constant7(self):
        """
        1 mol/kg LiCl = 64
        """
        s1 = pyEQL.Solution([["Li+", "1 mol/kg"], ["Cl-", "1 mol/kg"]])

        result = s1.dielectric_constant.magnitude
        expected = 64

        assert np.isclose(result, expected, rtol=RTOL)

    @pytest.mark.xfail()
    def test_dielectric_constant8(self):
        """
        12 mol/kg LiCl = 24
        """
        s1 = pyEQL.Solution([["Li+", "12 mol/kg"], ["Cl-", "12 mol/kg"]])

        result = s1.dielectric_constant.magnitude
        expected = 24

        assert np.isclose(result, expected, rtol=RTOL)

    def test_dielectric_constant9(self):
        """
        6.5 mol/kg RbCl = 43
        """
        s1 = pyEQL.Solution([["Rb+", "6.5 mol/kg"], ["Cl-", "6.5 mol/kg"]])

        result = s1.dielectric_constant.magnitude
        expected = 43

        assert np.isclose(result, expected, rtol=RTOL)

    def test_dielectric_constant10(self):
        """
        2.1 mol/kg RbCl = 59
        """
        s1 = pyEQL.Solution([["Rb+", "2.1 mol/kg"], ["Cl-", "2.1 mol/kg"]])

        result = s1.dielectric_constant.magnitude
        expected = 59

        assert np.isclose(result, expected, rtol=RTOL)

    def test_dielectric_constant11(self):
        """
        0.5 mol/kg RbCl = 73
        """
        s1 = pyEQL.Solution([["Rb+", "0.5 mol/kg"], ["Cl-", "0.5 mol/kg"]])

        result = s1.dielectric_constant.magnitude
        expected = 73

        assert np.isclose(result, expected, rtol=RTOL)



"""
pyEQL salt matching test suite
==============================

This file contains tests for the salt-matching algorithm used by pyEQL in
salt_ion_match.py
"""
import platform

import numpy as np
import pytest

import pyEQL
from pyEQL.salt_ion_match import Salt


def test_salt_init():
    s = Salt("Na[+1]", "Cl[-1]")
    assert s.formula == "NaCl"
    assert s.cation == "Na[+1]"
    assert s.anion == "Cl[-1]"
    assert s.nu_anion == 1
    assert s.nu_cation == 1
    assert s.z_cation == 1
    assert s.z_anion == -1

    s = Salt("Fe+3", "OH-1")
    assert s.formula == "Fe(OH)3"
    assert s.cation == "Fe[+3]"
    assert s.anion == "OH[-1]"
    assert s.nu_anion == 3
    assert s.nu_cation == 1
    assert s.z_cation == 3
    assert s.z_anion == -1


def test_empty_solution():
    """
    test matching a solution that contains no solutes other than water
    """
    s1 = pyEQL.Solution()
    assert isinstance(s1.get_salt(), pyEQL.salt_ion_match.Salt)
    assert s1.get_salt().formula == "HOH"
    assert s1.get_salt().cation == "H[+1]"
    assert s1.get_salt().anion == "OH[-1]"
    assert s1.get_salt().nu_cation == 1
    assert s1.get_salt().nu_anion == 1


def test_single_salt_mono():
    """
    test matching a solution with a single monovalent salt
    """
    s1 = pyEQL.Solution([["Na+", "2 mol/L"], ["Cl-", "2 mol/L"]])
    assert isinstance(s1.get_salt(), pyEQL.salt_ion_match.Salt)
    assert s1.get_salt().formula == "NaCl"
    assert s1.get_salt().cation == "Na[+1]"
    assert s1.get_salt().anion == "Cl[-1]"
    assert s1.get_salt().nu_cation == 1
    assert s1.get_salt().nu_anion == 1


def test_single_salt_di():
    """
    test matching a solution with a single divalent salt
    """
    s1 = pyEQL.Solution([["Na+", "4 mol/L"], ["SO4-2", "2 mol/L"]])
    assert isinstance(s1.get_salt(), pyEQL.salt_ion_match.Salt)
    assert s1.get_salt().formula == "Na2SO4"
    assert s1.get_salt().cation == "Na[+1]"
    assert s1.get_salt().anion == "SO4[-2]"
    assert s1.get_salt().nu_cation == 2
    assert s1.get_salt().nu_anion == 1


def test_single_salt_di2():
    """
    test matching a solution with a single divalent salt
    """
    s1 = pyEQL.Solution([["Fe+3", "1 mol/L"], ["Cl-", "3 mol/L"]])
    assert isinstance(s1.get_salt(), pyEQL.salt_ion_match.Salt)
    assert s1.get_salt().formula == "FeCl3"
    assert s1.get_salt().cation == "Fe[+3]"
    assert s1.get_salt().anion == "Cl[-1]"
    assert s1.get_salt().nu_cation == 1
    assert s1.get_salt().nu_anion == 3


def test_single_ion():
    """
    test matching a solution containing only a single ion
    """
    s1 = pyEQL.Solution([["Fe+3", "1 mol/L"]])
    assert isinstance(s1.get_salt(), pyEQL.salt_ion_match.Salt)
    assert s1.get_salt().formula == "Fe(OH)3"
    assert s1.get_salt().cation == "Fe[+3]"
    assert s1.get_salt().anion == "OH[-1]"
    assert s1.get_salt().nu_cation == 1
    assert s1.get_salt().nu_anion == 3


@pytest.mark.skipif(platform.machine() == "arm64" and platform.system() == "Darwin", reason="arm64 not supported")
def test_salt_with_equilibration():
    """
    test matching a solution containing a salt, before and after equilibration.
    Due to speciation changes, the concentration of the salt will decrease unless
    get_salt_dict() uses total concentrations
    """
    s1 = pyEQL.Solution({"Mg+2": "1 mol/L", "Cl-": "2 mol/L"})
    assert isinstance(s1.get_salt(), pyEQL.salt_ion_match.Salt)
    assert s1.get_salt().formula == "MgCl2"
    assert s1.get_salt().cation == "Mg[+2]"
    assert s1.get_salt().anion == "Cl[-1]"
    assert s1.get_salt().nu_cation == 1
    assert s1.get_salt().nu_anion == 2
    assert np.isclose(s1.get_salt_dict()["MgCl2"]["mol"], 1)
    s1.equilibrate()
    assert s1.get_salt().formula == "MgCl2"
    assert np.isclose(s1.get_salt_dict()["MgCl2"]["mol"], 1)


def test_salt_asymmetric():
    """
    test matching a solution where the cation and anion concentrations
    are not equal
    """
    s1 = pyEQL.Solution([["Na+", "1 mol/kg"], ["Cl-", "4 mol/kg"]])
    assert isinstance(s1.get_salt(), pyEQL.salt_ion_match.Salt)
    assert s1.get_salt().formula == "NaCl"
    assert s1.get_salt().cation == "Na[+1]"
    assert s1.get_salt().anion == "Cl[-1]"
    assert s1.get_salt().nu_cation == 1
    assert s1.get_salt().nu_anion == 1



"""
pyEQL test suite for bulk property calculations
===============================================

This file contains tests for the bulk properties calculated by
Solution class methods. Currently included methods are:

- hardness

"""


import pyEQL


# an empty solution should have zero hardness
def test_empty_solution():
    s1 = pyEQL.Solution()
    result = s1.hardness.magnitude
    expected = 0

    assert result == expected


# a solution with only monovalent ions should have zero hardness
def test_hardness_1():
    s1 = pyEQL.Solution([["Na+", "0.2 mol/L"], ["Cl-", "0.2 mol/L"]])
    result = s1.hardness.magnitude
    expected = 0

    assert result == expected


# a solution with only multivalent anions should have zero hardness
def test_hardness_2():
    s1 = pyEQL.Solution([["Na+", "0.4 mol/L"], ["SO4-2", "0.2 mol/L"]])
    result = s1.hardness.magnitude
    expected = 0

    assert result == expected


# the hardness should return the equivalent concentration, not just the
# molar concentration (e.g. multiply by the charge)
def test_hardness_3():
    s1 = pyEQL.Solution([["Fe+3", "0.1 mol/L"], ["Cl-", "0.3 mol/L"]])
    result = s1.hardness.magnitude
    expected = 15013.5

    assert round(abs(result - expected), 7) == 0


# the hardness should account for multiple cations but count only those
# that are multivalent
def test_hardness_4():
    s1 = pyEQL.Solution(
        [
            ["Na+", "0.1 mol/L"],
            ["K+", "0.1 mol/L"],
            ["Mg+2", "0.1 mol/L"],
            ["Ca+2", "0.1 mol/L"],
            ["Fe+3", "0.1 mol/L"],
            ["Cl-", "0.1 mol/L"],
            ["F-", "0.1 mol/L"],
            ["SO4-2", "0.2 mol/L"],
            ["PO4-3", "0.1 mol/L"],
        ]
    )
    result = s1.hardness.magnitude
    expected = 35031.5

    assert round(abs(result - expected), 7) == 0


# the hardness should return g/L units
def test_hardness_5():
    s1 = pyEQL.Solution([["Fe+3", "0.1 mol/L"], ["Cl-", "0.3 mol/L"]])
    result = str(s1.hardness.dimensionality)
    expected = "[mass] / [length] ** 3"

    assert result == expected
